#include "../common/pch.h"


// constructor
DTSS::DTSS(float maxValue, int set_cardinality, float node_prob, char* dataset_file_name, int TO_dimension, int PO_dimension, int data_cardinality, char dataset_type){
	this->maxValue = maxValue;
	this->set_cardinality = set_cardinality;
	this->node_prob = node_prob;
	this->dataset_file_name = dataset_file_name;
	this->TO_dimension = TO_dimension;
	this->PO_dimension = PO_dimension;
	this->data_cardinality = data_cardinality;
	this->dataset_type = dataset_type;

	scaledDataFileName = getScaledFileName(dataset_file_name, set_cardinality, node_prob, TO_dimension, PO_dimension);
	scaledDataFileNamePath = getScaledFileNamePath(dataset_file_name, set_cardinality, node_prob, TO_dimension, PO_dimension);
	outDirectory = getOutDirectory(dataset_file_name, "DTSS", scaledDataFileName);

	// initialize statistics
	numResults = 0;
	accesses = 0;

	initialize();
}

DTSS::~DTSS(){
	for(int i=0; i<PO_dimension; i++){
		delete G[i];
	}
	delete[] G;
}

void DTSS::initialize(){
	int *graphSizes = new int[PO_dimension];
	G = new Graph*[PO_dimension];
	for(int i=0; i<PO_dimension; i++){
		// Initialization of each graph and its ST
		G[i] = new Graph(set_cardinality, node_prob);
		G[i]->createST();
		G[i]->computePODominanceClosure();
		graphSizes[i] = G[i]->graph_size;
	}
	initiateScaledFile(dataset_file_name, maxValue, set_cardinality, node_prob, TO_dimension, PO_dimension, graphSizes);
	delete[] graphSizes;
}


void DTSS::execute(){
	// file for calculating progressiveness
	string progrFileName = outDirectory + scaledDataFileName + "_DTSS_progr";
	progr.open(progrFileName.c_str(), ios::out);

	string scaledDataFileNamePath = outDirectory + scaledDataFileName;
	ifstream dataset_file;
	dataset_file.open(scaledDataFileNamePath.c_str());

	Timing *totim = new Timing();
	totim->startClock();


	// group creation for DTSS+
	ostringstream dictFile;
	dictFile << outDirectory << scaledDataFileName << "_DTSS_dict";
	cerr << "dictFile = " << dictFile.str() << endl;

	ifstream inp;
	inp.open(dictFile.str().c_str(), ifstream::in);
	inp.close();
	if(inp.fail()){
		inp.clear(ios::failbit);
		cerr << "gonna create groups" << endl;
		createSubDatasets();
		cerr << "groups created" << endl;
		ofstream dict;
		dict.open(dictFile.str().c_str(), ofstream::out);

		multimap<int, string>::iterator ki;
		for (ki = dictionary.begin(); ki!=dictionary.end(); ki++){
			string dagroup = ki->second;
			dict << dagroup << " " << groupSizes[dagroup];
			dict << endl;
		}
		dict.close();
	}
	else{
		cerr << "groups file exists" << endl;
		inp.open(dictFile.str().c_str(), ifstream::in);
		string group;
		int theGroupSize;

		while(!inp.eof()){
			inp >> group >> theGroupSize;
			if (inp.eof()) break;
			dictionary.insert(make_pair(computeRectDist(group), group));
			groupSizes.insert(make_pair(group, theGroupSize));
		}

		inp.close();

		cerr << dictionary.size() << " groups read from dictionary" << endl;
	}

	RT = new DTSS_RTree*[dictionary.size()];
	createRTrees();

	cerr << "finished loading trees" << endl;

	// Skyline Query
	Timing *tim = new Timing();
	tim->startClock();
//	unsigned __int64 temp = pi.cpu_time_ms();
	skyline();
//	query_time = pi.cpu_time_ms() - temp;
	query_time = tim->stopClock();
	delete tim;
	cerr << "query done!" << endl;

	numTotalLeafNodes = 0;
	numTotalInternalNodes = 0;
//	cerr << "flushing R-trees to disk..." << endl;
	for (int i=0; i<dictionary.size(); i++){
		RT[i] = new DTSS_RTree(this, treeNames[i].c_str(), NULL);
		numTotalLeafNodes += RT[i]->num_of_dnodes;
		numTotalInternalNodes += RT[i]->num_of_inodes;
		delete RT[i];
	}
	delete[] RT;
	total_time = totim->stopClock();
	delete totim;
	progr.close();
}


void DTSS::createSubDatasetsOLD(){
	// read from file
	ifstream dataset_file;
	dataset_file.open(scaledDataFileNamePath.c_str());


	float *PO_point = new float[PO_dimension];
	float *TO_point = new float[TO_dimension];
	float *point = new float[TO_dimension]; // changed from deprecated code


	map<string, ofstream*> fileHandles; // <group_id, pointer to file>

	// First split the dataset into separate files according to their
	while(!dataset_file.eof()){

		// Read the PO attribute values of a point point
		dataset_file >> TO_point[0];    // This is necessary in order not
		if(dataset_file.eof()) break;	// to insert the last point twice

		for(int i=1; i<TO_dimension; i++){
			dataset_file >> TO_point[i];
		}
		for(int i=0; i<PO_dimension; i++){
			dataset_file >> PO_point[i];
		}
		// Insert the point to the corresponding file
		string group = determineGroup(PO_point);
		if (fileHandles.find(group) == fileHandles.end()){ // new group
			groupSizes.insert(make_pair(group, 0)); // initialize count
			dictionary.insert(make_pair(computeRectDist(group), group));
			ostringstream dastream;
			dastream << outDirectory << scaledDataFileName << "_DTSS_grp" << group;
			string file_name = dastream.str();
			ofstream *foufou = new ofstream(file_name.c_str(), ios::out);
			fileHandles.insert(make_pair(group, foufou));
		}

		groupSizes[group]++;

		memcpy(point, TO_point, sizeof(float)*TO_dimension);
		insertIntoFile(fileHandles[group], point);
	}
	delete[] PO_point;
	delete[] TO_point;
	delete[] point;


	// close input file
	dataset_file.close();

	// close all ofstreams
	map<string, ofstream*>::iterator iter;
	for (iter=fileHandles.begin(); iter!=fileHandles.end(); iter++){
		iter->second->close();
	}

//	map<string, int>::iterator iti;
//	for (iti=groupSizes.begin(); iti!=groupSizes.end(); iti++){
//		cerr << "group " << iti->first << " has " << iti->second << " points" << endl;
//	}

}


void DTSS::createSubDatasets(){
	// read from file
	ifstream dataset_file;
	dataset_file.open(scaledDataFileNamePath.c_str());

	float *PO_point = new float[PO_dimension];
	float *TO_point = new float[TO_dimension];

	map<string, vector<float*> > buffer; // buffer for storing group points
	int bufSize = 1000; // buffer size per group

	// First split the dataset into separate files according to their
	while(!dataset_file.eof()){

		// Read the PO attribute values of a point point
		dataset_file >> TO_point[0];    // This is necessary in order not
		if(dataset_file.eof()) break;	// to insert the last point twice

		for(int i=1; i<TO_dimension; i++){
			dataset_file >> TO_point[i];
		}
		for(int i=0; i<PO_dimension; i++){
			dataset_file >> PO_point[i];
		}
		// Insert the point to the corresponding file
		string group = determineGroup(PO_point);
		if (buffer.find(group) == buffer.end()){ // new group
			groupSizes.insert(make_pair(group, 0)); // initialize count
			dictionary.insert(make_pair(computeRectDist(group), group));
		}

		float *point = new float[TO_dimension];
		memcpy(point, TO_point, sizeof(float)*TO_dimension);
		buffer[group].push_back(point);
		groupSizes[group]++;

		if (buffer[group].size() == bufSize){ // flush and delete the buffer
			flushIntoFile(group, buffer[group]);
		}

	}

	// flush and delete all buffers
	map<string, vector<float*> >::iterator iter;
	for (iter=buffer.begin(); iter!=buffer.end(); iter++){
		flushIntoFile(iter->first, iter->second);
	}

	delete[] PO_point;
	delete[] TO_point;

	// close input file
	dataset_file.close();


//	map<string, int>::iterator iti;
//	for (iti=groupSizes.begin(); iti!=groupSizes.end(); iti++){
//		cerr << "group " << iti->first << " has " << iti->second << " points" << endl;
//	}

}


void DTSS::createRTrees(){
	int dimension = TO_dimension;
	multimap<int, string>::iterator it;

	int k=0;
//	FILE *fp;
	for(it=dictionary.begin(); it!=dictionary.end(); ++it){
		// create file names
		ostringstream dastream;
		dastream << outDirectory << scaledDataFileName << "_DTSS_grp" << it->second;
		string dataset_file_name = dastream.str();
		string tree_file_name = dastream.str();
		tree_file_name.append(".rtree");

		treeNames.insert(make_pair(k, tree_file_name));

//		cerr << "k = " << k << endl;
//		fp = fopen(dataset_file_name.c_str(), "r");
//		if (!fp){
//			cerr << "test could not open file " << dataset_file_name << endl;
//		}
////		fclose(fp);


		// if the corresponding tree does not exist, create it
		ifstream fin(tree_file_name.c_str(), ios::in);
		if (fin.fail()){ // create the trees
			RT[k] = new DTSS_RTree(this, tree_file_name.c_str(), 4096, NULL, dimension);
//			RT[k]->bulkloadDUMMY(fp, dataset_file_name.c_str(), dimension, groupSizes[it->second], 15);
			RT[k]->bulkload(dataset_file_name.c_str(), dimension, groupSizes[it->second], 15);
		}
		else { // restore tree from file
			RT[k] = new DTSS_RTree(this, tree_file_name.c_str(), NULL);
		}
		delete RT[k];
//		fclose(fp);
		fin.close();

		k++;
	}
}


void DTSS::flushIntoFile(string group, vector<float *>& points){
	ostringstream dastream;
	dastream << outDirectory << scaledDataFileName << "_DTSS_grp" << group;
	string file_name = dastream.str();
	ofstream foufou(file_name.c_str(), ios::app);
	vector<float *>::iterator iter;
	for (iter=points.begin(); iter!=points.end(); iter++){
		float* point = *iter;
		for(int i=0;i<TO_dimension;i++){
			foufou << point[i];
			if (i == TO_dimension-1) foufou << endl;
			if (i != TO_dimension-1) foufou << " ";
		}
		delete[] point; // empty buffer
	}
	points.clear();
	foufou.close();
}

void DTSS::insertIntoFile(ofstream *fileout, float* point){
	for(int i=0;i<TO_dimension;i++){
		(*fileout) << point[i];
		if(i == TO_dimension-1)
			(*fileout) << endl;
		if (i != TO_dimension-1)
			(*fileout) << " ";
	}
}


void DTSS::skyline(){

	multimap<int, string>::iterator iter;
	int k = 0;
	for(iter=dictionary.begin(); iter!=dictionary.end(); iter++){
		float* PO_values = new float[PO_dimension]; // PO values extracted from the group identifier
		assignStringToNum(PO_values, iter->second, PO_dimension);
		XxkHeap *hp = new XxkHeap();

		RT[k] = new DTSS_RTree(this, treeNames[k].c_str(), NULL);
		RT[k]->skyline(hp, PO_values);
		delete RT[k];
		delete hp;
		delete[] PO_values;
		k++;
	}
	numResults = result.size()/(TO_dimension+PO_dimension);
}


string DTSS::determineGroup(float* PO_point){
	stringstream ss;
	for(int i=0;i<PO_dimension;i++){
		ss << PO_point[i];
		if(i!=PO_dimension-1)
			ss << "_";
	}
	return ss.str();
}

void DTSS::assignStringToNum(float* n, string s, int dim){
	stringstream ss(s);
	char c;
	for(int i=0;i<dim;i++){
		ss >> n[i];
		if(i!=dim-1)
			ss >> c;
	}
}

int DTSS::computeRectDist(string group){
	stringstream ss(group);
	int dist = 0, coord;
	char c;
	for(int i=0;i<PO_dimension;i++){
		ss >> coord;
		if(i!=PO_dimension-1)
			ss >> c;
		dist += coord;
	}
	return dist;
}

/* Old PO_dominate)
bool DTSS::PO_dominate(float* _pt1, float* _pt2){ // _pt1 is a point, _pt2 is an MBR
	int encoding1, encoding2, index1, index21, index22;

	for (int i =0; i < PO_dimension; i++){
		index1 =  (int) _pt1[i];
		encoding1 = G[i]->TO_table[index1];
		//Calculate the encoding of the MBR represented by _pt2 on the PO dimensions
		index21 = (int) _pt2[2*i];
		index22 = (int) _pt2[2*i+1];
		encoding2 = 0;

		for(int j=index21;j<=index22;j++){
			encoding2 = (encoding2 | G[i]->TO_table[j]);
		}

		//Check for dominance
//		if(encoding1 == encoding2) return false;

		if((encoding1 | encoding2) != encoding1)
			return false;
	}
	return true;
}
*/

bool DTSS::PO_dominate(float* _pt1, float* _pt2){ // _pt1 is a point, _pt2 is an MBR
	int encoding1, encoding2, i, j;

	for (i =0; i < PO_dimension; i++){
		encoding1 = G[i]->TO_table[(int) _pt1[i]];

		for(j=(int) _pt2[2*i]; j<=(int) _pt2[2*i+1]; j++){
			encoding2 = G[i]->TO_table[j];
			if(encoding2 != -1 && (G[i]->nodes_table[encoding1].PODominates[encoding2][0]!=1))
				return false;
		}
	}

	return true;
}


string DTSS::getStatistics(){
	ostringstream sout;

	sout << "----------------- DTSS ----------------" << endl;
	sout << ":: Groups Info ::" << endl;
	sout << "Number of groups = " << groupSizes.size() << "." << endl;
	int count = 0;
	map<string, int>::iterator it;
//	sout << "(group : num of points) :: ";
	for (it=groupSizes.begin(); it!=groupSizes.end(); it++){
		count += it->second;
//		sout << "(" << it->first << ":" << it->second << ") ";
	}
//	sout << endl;
	sout << "Total Points " << count << endl;
	sout << "Total (leaf nodes, internal nodes) in all R-Trees = (" << numTotalLeafNodes << ", " << numTotalInternalNodes << ")" << endl;

	sout << ":: Skyline Info ::" << endl;
	sout << "Number of Results = " << numResults << endl;
	sout << "Node accesses = " << accesses << endl;
	sout << "Query CPU time = " << query_time << endl;
	sout << "Total CPU time = " << total_time << endl;
	sout << "---------------------------------------" << endl;

	return sout.str();
}

void DTSS::printResults(){
	int dim = TO_dimension+PO_dimension;
	vector<float>::iterator iti = result.begin();
	while (iti != result.end()){
		for(int j=0;j<dim;j++){
			cerr << *iti << " ";
			iti++;
		}
		cerr << endl;
	}
	cerr<<result.size()/dim<<endl;
}
