#include "../common/pch.h"


// constructor
SDC::SDC(float maxValue, int set_cardinality, float node_prob, char* dataset_file_name, int TO_dimension, int PO_dimension, int data_cardinality, char dataset_type){
	this->maxValue = maxValue;
	this->set_cardinality = set_cardinality;
	this->node_prob = node_prob;
	this->dataset_file_name = dataset_file_name;
	this->TO_dimension = TO_dimension;
	this->PO_dimension = PO_dimension;
	this->data_cardinality = data_cardinality;
	this->dataset_type = dataset_type;

	scaledDataFileName = getScaledFileName(dataset_file_name, set_cardinality, node_prob, TO_dimension, PO_dimension);
	scaledDataFileNamePath = getScaledFileNamePath(dataset_file_name, set_cardinality, node_prob, TO_dimension, PO_dimension);
	outDirectory = getOutDirectory(dataset_file_name, "SDC", scaledDataFileName);

	// initialize statistics
	numResults = 0;
	false_hits = 0;
	accesses = 0;

	initialize();
}

SDC::~SDC(){
	for(int i=0; i<PO_dimension; i++){
		delete G[i];
	}
	delete[] G;
}

void SDC::initialize(){
	int *graphSizes = new int[PO_dimension];
	G = new Graph*[PO_dimension];
	for(int i=0; i<PO_dimension; i++){
		// Initialization of each graph and its ST
		G[i] = new Graph(set_cardinality, node_prob);
//		G[i] = new Graph("smallgraph.txt", node_prob);
		G[i]->createST();
		G[i]->computePODominanceClosure();
		// Initialization for SDC+ only
		G[i]->createIntervalsTable_SDC();
		G[i]->labelNodesStrata();

		graphSizes[i] = G[i]->graph_size;
	}
	initiateScaledFile(dataset_file_name, maxValue, set_cardinality, node_prob, TO_dimension, PO_dimension, graphSizes);
	delete[] graphSizes;
}


void SDC::execute(){
	// file for calculating progressiveness
	string progrFileName = outDirectory + scaledDataFileName + "_SDC_progr";
	progr.open(progrFileName.c_str(), ios::out);


	// strata creation for SDC+ only
	ostringstream dictFile;
	dictFile << outDirectory << scaledDataFileName << "_SDC_dict";
	cerr << "dictFile = " << dictFile.str() << endl;


	totim = new Timing();
	totim->startClock();

	ifstream inp;
	inp.open(dictFile.str().c_str(), ifstream::in);
	inp.close();
	if(inp.fail()){
		inp.clear(ios::failbit);
		cerr << "gonna create strata" << endl;
		createSubDatasets();
		cerr << "strata created" << endl;
		ofstream dict;
		dict.open(dictFile.str().c_str(), ofstream::out);

		for (int kl=0; kl<dictionary.size(); kl++){
			int dastratum = dictionary[kl];
			dict << dastratum << " " << strataSizes[dastratum];
			dict << endl;
		}
		dict.close();
	}
	else{
		cerr << "strata file exists" << endl;
		inp.open(dictFile.str().c_str(), ifstream::in);
		int stratum, theStratumSize;

		while(!inp.eof()){
			inp >> stratum >> theStratumSize;
			if (inp.eof()) break;
			dictionary.push_back(stratum);
			strataSizes.insert(make_pair(stratum, theStratumSize));
		}

		inp.close();

		cerr << dictionary.size() << " strata read from dictionary" << endl;
		for (int kl=0; kl<dictionary.size(); kl++){
			int stratum = dictionary[kl];
		}

	}

	RT = new SDC_RTree*[dictionary.size()];
	std::sort(dictionary.begin(), dictionary.end()); // for vector
	createRTrees();

	cerr << "finished loading trees" << endl;

	// Skyline Query
	Timing *tim = new Timing();
	tim->startClock();
//	unsigned __int64 temp = pi.cpu_time_ms();
	skyline();
//	query_time = pi.cpu_time_ms() - temp;
//	cerr << "pi " << pi.cpu_time_ms() - temp << endl;
	query_time = tim->stopClock();
	delete tim;
//	cerr << "ti " << query_time << endl;
	cerr << "query done!" << endl;

	numTotalLeafNodes = 0;
	numTotalInternalNodes = 0;
//	cerr << "flushing R-trees to disk..." << endl;
	for (int i=0; i<dictionary.size(); i++){
		numTotalLeafNodes += RT[i]->num_of_dnodes;
		numTotalInternalNodes += RT[i]->num_of_inodes;
		delete RT[i];
	}
	delete[] RT;
	total_time = totim->stopClock();
	delete totim;
	progr.close();
}


void SDC::createSubDatasets(){
	// read from file
	ifstream dataset_file;
	dataset_file.open(scaledDataFileNamePath.c_str());


	float *PO_point = new float[PO_dimension];
	float *TO_point = new float[TO_dimension];
	float *point = new float[TO_dimension + PO_dimension];


	map<int, ofstream*> fileHandles; // <stratum_id, pointer to file>

	// First split the dataset into separate files according to their
	while(!dataset_file.eof()){

		// Read the PO attribute values of a point point
		dataset_file >> TO_point[0];    // This is necessary in order not
		if(dataset_file.eof()) break;	// to insert the last point twice

		for(int i=1; i<TO_dimension; i++){
			dataset_file >> TO_point[i];
		}
		for(int i=0; i<PO_dimension; i++){
			dataset_file >> PO_point[i];
		}
		// Insert the point to the corresponding file
		int stratum = determineStratum(PO_point);
		if (fileHandles.find(stratum) == fileHandles.end()){ // new stratum
			strataSizes.insert(make_pair(stratum, 0)); // initialize count
			dictionary.push_back(stratum);
			ostringstream dastream;
			dastream << outDirectory << scaledDataFileName << "_SDC_str" << stratum;
			string file_name = dastream.str();
			ofstream *foufou = new ofstream(file_name.c_str(), ios::out);
			fileHandles.insert(make_pair(stratum, foufou));
		}

		strataSizes[stratum]++;

		memcpy(point, TO_point, sizeof(float)*TO_dimension);
		memcpy(&point[TO_dimension], PO_point, sizeof(float)*PO_dimension);
		insertIntoFile(fileHandles[stratum], point);
	}
	delete[] PO_point;
	delete[] TO_point;
	delete[] point;


	// close input file
	dataset_file.close();

	// close all ofstreams
	map<int, ofstream*>::iterator iter;
	for (iter=fileHandles.begin(); iter!=fileHandles.end(); iter++){
		iter->second->close();
	}

//	map<int, int>::iterator iti;
//	for (iti=strataSizes.begin(); iti!=strataSizes.end(); iti++){
//		cerr << "stratum " << iti->first << " has " << iti->second << " points" << endl;
//	}

}

void SDC::insertIntoFile(ofstream *fileout, float* point){
	int i;
	// I erased the dictionary;
	for(i=0;i<TO_dimension;i++){
		(*fileout) << point[i];
		if(i == TO_dimension-1 && PO_dimension == 0)
			(*fileout) << endl;
		if (i != TO_dimension-1)
			(*fileout) << " ";
		if (i == TO_dimension-1 && PO_dimension!=0)
			(*fileout) << " ";
	}
	if(PO_dimension !=0 ){
		float *p = convertPoint(point+TO_dimension);
		for(i=0;i<PO_dimension*2;i++){
			(*fileout) << p[i];
			if(i == PO_dimension*2-1)
				(*fileout) << endl;
			else
				(*fileout) << " ";
		}
		delete p;
	}
}

float* SDC::convertPoint(float* point){
	float *p = new float[PO_dimension*2];
	for(int i=0;i<PO_dimension;i++){
		int index = (int) (point[i]);
		int encoding = G[i]->TO_table[index];
		p[2*i] = G[i]->IT[encoding][0][0];
		p[2*i+1] = (G[i]->graph_size*2)-(G[i]->IT[encoding][0][1])+1;
	}
	return p;
}

int SDC::determineStratum(float *PO_point){
	int i, stratum, temp_covered;
	int covered = 0;
	bool covering = true;

	for(i=0;i<PO_dimension;i++){
		covering = covering && G[i]->covering(PO_point[i]);
		temp_covered = G[i]->covered(PO_point[i]);
		if(temp_covered > covered)
			covered = temp_covered;
	}

	stratum = covered * 10;
	if(covering)
		stratum += 1;

	return stratum;
}

void SDC::createRTrees(){
	int dimension = TO_dimension + 2*PO_dimension;
	vector<int>::iterator it;

	int k=0;
	for(it=dictionary.begin(); it!=dictionary.end(); ++it){
		// create file names
		ostringstream dastream;
		dastream << outDirectory << scaledDataFileName << "_SDC_str" << *it;
		string dataset_file_name = dastream.str();
		string tree_file_name = dastream.str();
		tree_file_name.append(".rtree");

		// if the corresponding tree does not exist, create it
		ifstream fin(tree_file_name.c_str(), ios::in);
		if (fin.fail()){ // create the trees
			RT[k] = new SDC_RTree(this, tree_file_name.c_str(), 4096, NULL, dimension);
			RT[k]->bulkload(dataset_file_name.c_str(), dimension, strataSizes[*it], 15);
			k++;
		}
		else { // restore tree from file
			RT[k] = new SDC_RTree(this, tree_file_name.c_str(), NULL);
			k++;
		}
		fin.close();
	}
}


void SDC::skyline(){
	int i, j, k, dim=TO_dimension+2*PO_dimension;

	for(i=0; i<dictionary.size(); i++){
		XxkHeap *hp = new XxkHeap();
		RT[i]->skyline(hp);
		mergeResults_iter_simple();
//		mergeResults_iter();
		delete hp;
	}
	numResults = global_result.size()/dim;
	//printResults(global_result, dim);
}


void SDC::mergeResults_iter_simple(){
	list<float>::iterator liter;
	int counter = 0;
	int dim = TO_dimension + 2*PO_dimension;
	for (liter=local_result.begin(); liter!=local_result.end(); liter++){
		global_result.push_back(*liter);
		counter++;
		if (counter%dim==0){
			progr << accesses << "\t" << totim->elapsed() << endl;
		}
	}
	local_result.clear();
}


void SDC::mergeResults_iter(){
	int _dim = TO_dimension+2*PO_dimension, i, j, k;
	bool ret = false;

	list<float>::iterator liter, giter;
	liter = local_result.begin();
	//cerr<<"Local: "<< local_result.size()<<endl;
	while (liter!=local_result.end()){
		float *lp = new float[_dim];
		for (j=0; j<_dim; j++){
			lp[j] = *liter;
			liter++;
		}

		if(global_result.empty()){
			for(j=0; j<_dim; j++){
				global_result.push_back(lp[j]);
			}
			numResults++;
		}
		else{
			giter=global_result.begin();
			while (giter!=global_result.end()){
				float *gp = new float[_dim];
				for (j=0; j<_dim; j++){
					gp[j] = *giter;
					giter++;
				}

				if(dominate(gp, lp)){
					ret = true;
					false_hits++;
					delete gp;
					break;
				}
				delete gp;
			}
			if(!ret){ // the point is not dominated
				progr << accesses << "\t" << totim->elapsed() << endl;
				for(k=0; k<_dim; k++){
					global_result.push_back(lp[k]);
				}
				numResults++;
			}
			ret = false;
		}
		delete lp;
	}
	local_result.clear();
}


bool SDC::m_dominate(int _dim, float *_pt1, float *_pt2){
	if(!TO_dominate(_dim, _pt1, _pt2))
		return false;

	return true;
}

/* Old dominate
bool SDC::dominate(float *_pt1, float *_pt2){
	int i;

	//TO atributes check
	if(!TO_dominate(TO_dimension, _pt1, _pt2))
		return false;

	//PO attributes check
	for (i =0; i < PO_dimension; i ++){

		// G[i]->invert_IT[_pt1[TO_dimension+2*i]]==G[i]->invert_IT[_pt2[TO_dimension+2*i]] ||
		if (  ((G[i]->invert_IT[(int) _pt1[TO_dimension+2*i]] | G[i]->invert_IT[(int) _pt2[TO_dimension+2*i]]) != G[i]->invert_IT[(int) _pt1[TO_dimension+2*i]])  ){
			return false;
		}
	}

	return true;
}
*/

bool SDC::dominate(float *_pt1, float *_pt2){
	//TO atributes check
	if(!TO_dominate(TO_dimension, _pt1, _pt2)){
		return false;
	}

	//PO attributes check
	int encoding1, encoding2;
	for (int i=0; i < PO_dimension; i++){
		encoding1 = G[i]->invert_IT[(int) _pt1[TO_dimension+2*i]];
		encoding2 = G[i]->invert_IT[(int) _pt2[TO_dimension+2*i]];
		if(G[i]->nodes_table[encoding1].PODominates[encoding2]!=1){
			return false;
		}
	}

	return true;
}

string SDC::getStatistics(){
	ostringstream sout;

	sout << "----------------- SDC -----------------" << endl;
	sout << ":: Strata Info ::" << endl;
	sout << "Number of strata = " << strataSizes.size() << "." << endl;
	map<int, int>::iterator it;
	sout << "(stratum : num of points) :: ";
	for (it=strataSizes.begin(); it!=strataSizes.end(); it++){
		sout << "(" << it->first << ":" << it->second << ") ";
	}
	sout << endl;
	sout << "Total (leaf nodes, internal nodes) in all R-Trees = (" << numTotalLeafNodes << ", " << numTotalInternalNodes << ")" << endl;

	sout << ":: Skyline Info ::" << endl;
	sout << "Number of Results = " << numResults << endl;
	sout << "Number of false hits in SDC = " << false_hits << endl;
	sout << "Node accesses = " << accesses << endl;
	sout << "Query CPU time = " << query_time << endl;
	sout << "Total CPU time = " << total_time << endl;
	sout << "---------------------------------------" << endl;

	return sout.str();
}

void SDC::printResults(){
	int dim = TO_dimension+2*PO_dimension;
	list<float>::iterator iti = global_result.begin();
	while (iti != global_result.end()){
		for(int j=0;j<dim;j++){
			cerr << *iti << " ";
			iti++;
		}
		cerr << endl;
	}
	cerr<<global_result.size()/dim<<endl;
}
