#include "../common/pch.h"


// constructor
TSS::TSS(float maxValue, int set_cardinality, float node_prob, char* dataset_file_name, int TO_dimension, int PO_dimension, int data_cardinality, char dataset_type){
	this->maxValue = maxValue;
	this->set_cardinality = set_cardinality;
	this->node_prob = node_prob;
	this->dataset_file_name = dataset_file_name;
	this->TO_dimension = TO_dimension;
	this->PO_dimension = PO_dimension;
	this->data_cardinality = data_cardinality;
	this->dataset_type = dataset_type;

	scaledDataFileName = getScaledFileName(dataset_file_name, set_cardinality, node_prob, TO_dimension, PO_dimension);
	scaledDataFileNamePath = getScaledFileNamePath(dataset_file_name, set_cardinality, node_prob, TO_dimension, PO_dimension);
	outDirectory = getOutDirectory(dataset_file_name, "TSS", scaledDataFileName);
	cerr << "scaledDataFileName = " << scaledDataFileName << endl;
	cerr << "scaledDataFileNamePath = " << scaledDataFileNamePath << endl;
	cerr << "outDirectory = " << outDirectory << endl;

	// initialize statistics
	numResults = 0;
	accesses = 0;

	initialize();
}

TSS::~TSS(){
	for(int i=0; i<PO_dimension; i++){
		delete G[i];
	}
	delete[] G;
}

void TSS::initialize(){
	int *graphSizes = new int[PO_dimension];
	G = new Graph*[PO_dimension];
	for(int i=0; i<PO_dimension; i++){
		// Initialization of each graph and its ST
		G[i] = new Graph(set_cardinality, node_prob);
		G[i]->createST();
		G[i]->computePODominanceClosure();
		G[i]->computeRangeDomnedByEnc();
		graphSizes[i] = G[i]->graph_size;
	}
	initiateScaledFile(dataset_file_name, maxValue, set_cardinality, node_prob, TO_dimension, PO_dimension, graphSizes);
	delete[] graphSizes;
}

void TSS::execute(){
	// file for calculating progressiveness
	string progrFileName = outDirectory + scaledDataFileName + "_TSS_progr";
	progr.open(progrFileName.c_str(), ios::out);

	string rtreefilename = outDirectory + scaledDataFileName + ".rtree";

	totim = new Timing();
	totim->startClock();

	ifstream fin(rtreefilename.c_str(), ios::in);
	if (fin.fail()){
		cerr << "creating R-tree" << endl;
		RT = new TSS_RTree(this, rtreefilename.c_str(), 4096, NULL, TO_dimension+PO_dimension);
		RT->bulkload(scaledDataFileNamePath.c_str(), TO_dimension+PO_dimension, data_cardinality, 15);
	}
	else{
		cerr << "loading R-tree" << endl;
		RT = new TSS_RTree(this, rtreefilename.c_str(), NULL);
	}
	fin.close();
	cerr << "R-tree in memory" << endl;


	// Skyline Query
	Timing *tim = new Timing();
	tim->startClock();
//	unsigned __int64 temp = pi.cpu_time_ms();
	skyline();
//	query_time = pi.cpu_time_ms() - temp;
	query_time = tim->stopClock();
	delete tim;
	cerr << "query done!" << endl;

	numLeafNodes = RT->num_of_dnodes;
	numInternalNodes = RT->num_of_inodes;

//	cerr << "flushing R-tree to disk..." << endl;
	delete RT;

	total_time = totim->stopClock();
	delete totim;
	progr.close();
}

void TSS::skyline(){
	XxkHeap *hp = new XxkHeap();
//	unsigned __int64 temp = pi.cpu_time_ms();
	RT->skyline(hp);
//	query_time = pi.cpu_time_ms() - temp;
	numResults = result.size()/(TO_dimension+PO_dimension);
	delete hp;
}

/* Old PO_dominate
bool TSS::PO_dominate(float* _pt1, float* _pt2){ // _pt1 is a point, _pt2 is an MBR
	int encoding1, encoding2, index1, index21, index22;

	for (int i =0; i < PO_dimension; i++){
		index1 =  (int) _pt1[i];
		encoding1 = G[i]->TO_table[index1];
		//Calculate the encoding of the MBR represented by _pt2 on the PO dimensions
		index21 = (int) _pt2[2*i];
		index22 = (int) _pt2[2*i+1];
		encoding2 = 0;

		for(int j=index21;j<=index22;j++){
			encoding2 = (encoding2 | G[i]->TO_table[j]);
		}

		//Check for dominance
		if((encoding1 | encoding2) != encoding1)
			return false;
	}
	return true;
}
*/

bool TSS::PO_dominateOLD(float* _pt1, float* _pt2){ // _pt1 is a point, _pt2 is an MBR
	int encoding1, encoding2, i, j;

	for (i=0; i < PO_dimension; i++){
		encoding1 = G[i]->TO_table[(int) _pt1[i]];

		for(j=(int) _pt2[2*i]; j<=(int) _pt2[2*i+1]; j++){
			encoding2 = G[i]->TO_table[j];
			if (encoding2 == -1) {
				cerr << "**** TSS encoding2 is -1 ****" << endl;
			}
			if (G[i]->nodes_table[encoding2].encoding == -1) {
				cerr << "**** TSS real encoding2 is -1 ****" << endl;
			}
//			if(encoding2 != -1 && (G[i]->nodes_table[encoding1].PODominates[encoding2]!=1)){
			if(G[i]->nodes_table[encoding2].encoding != -1 && (G[i]->nodes_table[encoding1].PODominates[encoding2]!=1)){
				return false;
			}
		}
	}
	return true;
}


bool TSS::PO_dominate(float* _pt1, float* _pt2){ // _pt1 is a point, _pt2 is an MBR

	for (int i=0; i < PO_dimension; i++){
		int encoding = G[i]->TO_table[(int) _pt1[i]];

		int s = (int) _pt2[2*i];
		int e = (int) _pt2[2*i+1];

//		cerr << "TSS::PO_dominate checks TO range " <<

		if ( !G[i]->rangeDomnedByEnc[s][e][encoding] ){
			return false;
		}
	}
	return true;
}


string TSS::getStatistics(){
	ostringstream sout;

	sout << "----------------- TSS -----------------" << endl;
	sout << ":: R-Tree Info ::" << endl;
	sout << "(leaf nodes, internal nodes) in R-Tree = (" << numLeafNodes << ", " << numInternalNodes << ")" << endl;

	sout << ":: Skyline Info ::" << endl;
	sout << "Number of Results = " << numResults << endl;
	sout << "Node accesses = " << accesses << endl;
	sout << "Query CPU time = " << query_time << endl;
	sout << "Total CPU time = " << total_time << endl;
	sout << "---------------------------------------" << endl;

	return sout.str();
}

void TSS::printResults(){
	int dim = TO_dimension+PO_dimension;
	vector<float>::iterator iti = result.begin();
	while (iti != result.end()){
		for(int j=0;j<dim;j++){
			cerr << *iti << " ";
			iti++;
		}
		cerr << endl;
	}
	cerr<<result.size()/dim<<endl;
}

