#include "../common/pch.h"

Graph::Graph(int sc, float ndp){

	encoding_size = sc;
	lattice_size = 1 << sc;
	node_prob = ndp;

	createLattice();
	createFinalGraph();

	// Create table for fast reference of the topological order of the nodes
	TO_table = new int[graph_size];
}

Graph::Graph(char* fileName, float ndp){ // read graph from file... do not run createLattice!
	node_prob = ndp;
	encoding_size = -1;

	ifstream fin;
	fin.open(fileName, ifstream::in);
	fin >> lattice_size; // numNodes
//	cerr << "lattice_size = " << lattice_size << endl;
	nodes_table = new Node[lattice_size];
	for (int i=0; i<lattice_size; i++){
		nodes_table[i].init(encoding_size, i, this);
	}
	for (int i=0; i<lattice_size; i++){
		int numChildren;
		fin >> numChildren; // number of node's children
		for (int j=0; j<numChildren; j++){
			int id;
			fin >> id; // child id (from 0 to numNodes-1)
			nodes_table[i].out.push_back(id);
			nodes_table[id].in.push_back(i);
		}
	}
	fin.close();

	createFinalGraph();
	// Create table for fast reference of the topological order of the nodes
	TO_table = new int[graph_size];
//	cerr << "graph_size = " << graph_size << endl;
}


void Graph::createLattice(){

	nodes_table = new Node[lattice_size];

	for(int i=0;i<lattice_size;i++){
		nodes_table[i].init(encoding_size, i, this);
		nodes_table[i].setEdges();
	}
}

void Graph::createFinalGraph(){
	// Remove nodes. The root will not be deleted
//	srand(time(NULL));
	srand(12345);
	float p;
	int k=0;
	int deleted=0;
	// Create the list of deleted nodes
	for(int i=0; i<lattice_size; i++){
		p = (rand() % 10000 + 1) / float(10000);
		if((p > node_prob) && nodes_table[i].canBeDeleted()){
			nodes_table[i].eraseEdges();  // Delete the edges from and to the deleted node
			nodes_table[i].encoding = -1; // to indicate that the node with this encoding does not exist
			deleted++;
		}
	}
	graph_size = lattice_size-deleted;
}

void Graph::topologicalOrder(){
	int counter = 1;
	nodes_table[lattice_size-1].topologicalOrder(counter); // start from the root
}

void Graph::createST(){
	topologicalOrder();

	// examine nodes according to their topological order
	for(int i=0;i<graph_size;i++){
		nodes_table[TO_table[i]].assignSTParent();
		nodes_table[TO_table[i]].updatePredSet();
	}
}

void Graph::createIntervalsTable_SDC(){
	IT = new vector<vector<int> >[lattice_size];
	int counter = 0;
	nodes_table[lattice_size-1].assignTableIntervals(counter); // start from the root
}

void Graph::labelNodesStrata(){
	int i;
	for(i=0;i<graph_size;i++)
		nodes_table[TO_table[i]].assignCoveredInfo();
	for(i=graph_size-1;i>=0;i--)
		nodes_table[TO_table[i]].assignCoveringInfo();
}

bool Graph::covering(float point){
	//int TO = int(point)%graph_size;
	int TO = int(point);
	int encoding = TO_table[TO];
	return nodes_table[encoding].covering;
}

int Graph::covered(float point){
	//int TO = int(point)%graph_size;
	int TO = int(point);
	int encoding = TO_table[TO];
	return nodes_table[encoding].covered;
}

void Graph::computePODominanceClosure(){
	for(int i=graph_size-1;i>=0;i--){
		nodes_table[TO_table[i]].assignPODominanceInfo(graph_size);
	}
}


void Graph::computeRangeDomnedByEnc(){
//	cerr << "computing rangeDomnedByEnc" << endl;
	rangeDomnedByEnc = new bool**[graph_size];
	for (int s=0; s<graph_size; s++){
		rangeDomnedByEnc[s] = new bool*[graph_size];
		for (int e=s; e<graph_size; e++){
			rangeDomnedByEnc[s][e] = new bool[lattice_size];
			for (int enc=0; enc<lattice_size; enc++){ // for each encoding
				bool domned = true;
				for (int r=s; r<=e; r++){ // from s to e in topological order
					int rEnc = TO_table[r]; // get encoding
					if (nodes_table[enc].PODominates[rEnc]!=1){ // if encoding does not dominate the encoding of a value in the range
						domned = false;
						break;
					}
				}
				rangeDomnedByEnc[s][e][enc] = domned;
			}
		}
	}


//	for (int s=0; s<graph_size; s++){
//		for (int e=s; e<graph_size; e++){
//			for (int enc=0; enc<lattice_size; enc++){ // for each encoding
//				cerr << "s=" << s << " e=" << e << " enc=" << enc << " rangeDomnedByEnc=" << rangeDomnedByEnc[s][e][enc] << endl;
//			}
//		}
//	}
}
